import logging

import concurrent

from openstack.connection import Connection
from openstack.identity.v3.project import Project as OSProject
from openstack.load_balancer.v2.load_balancer import LoadBalancer as OSLoadBalancer

from osi_dump.importer.project.project_importer import ProjectImporter
from osi_dump.model.project import Project
import osi_dump.api.octavia as octavia_api
logger = logging.getLogger(__name__)


class OpenStackProjectImporter(ProjectImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_projects(self) -> list[Project]:
        """Import projects information from Openstack

        Raises:
            Exception: Raises exception if fetching project failed

        Returns:
            list[Instance]: _description_
        """

        logger.info(f"Importing projects for {self.connection.auth['auth_url']}")
        try:
            osload_balancers: list[OSLoadBalancer] = octavia_api.get_load_balancers(
                connection=self.connection
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch load_balancers for {self.connection.auth['auth_url']} {e}"
            ) from e

        project_lb_dict = {}

        for osload_balancer in osload_balancers: 
            if project_lb_dict.get(osload_balancer["project_id"]):
                project_lb_dict[osload_balancer["project_id"]] += 1 
            else: 
                project_lb_dict[osload_balancer["project_id"]] = 1

        try:
            osprojects: list[OSProject] = list(self.connection.identity.projects())
        except Exception as e:
            raise Exception(
                f"Can not fetch projects for {self.connection.auth['auth_url']}"
            ) from e

        projects: list[Project] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_project_info, project, project_lb_dict)
                for project in osprojects
            ]
            for future in concurrent.futures.as_completed(futures):
                projects.append(future.result())

        logger.info(f"Imported projects for {self.connection.auth['auth_url']}")

        return projects

    def _get_project_info(self, project: OSProject, project_lb_dict: dict) -> Project:
        
        usage_instance=None
        quota_instance=None
        usage_ram=None
        quota_ram=None
        usage_vcpu=None
        quota_vcpu=None

        try:
            compute_quotas = self.connection.compute.get_quota_set(
                project.id, usage=True
            )

            usage_instance=compute_quotas.usage["instances"]
            quota_instance=compute_quotas.instances
            usage_ram=compute_quotas.usage["ram"]
            quota_ram=compute_quotas.ram
            usage_vcpu=compute_quotas.usage["cores"]
            quota_vcpu=compute_quotas.cores
        except Exception as e:
            logger.warning(f"Get compute quotas failed for {project.id} error: {e}")

        usage_volume=None
        quota_volume=None
        usage_snapshot=None
        quota_snapshot=None
        usage_storage=None
        quota_storage=None
        try:
            storage_quotas = self.connection.block_storage.get_quota_set(
                project.id, usage=True
            )
            usage_volume=storage_quotas.usage["volumes"]
            quota_volume=storage_quotas.volumes
            usage_snapshot=storage_quotas.usage["snapshots"]
            quota_snapshot=storage_quotas.snapshots
            usage_storage=storage_quotas.usage["gigabytes"]
            quota_storage=storage_quotas.gigabytes
        except Exception as e:
            logger.warning(f"Get storage quotas failed for {project.id} error: {e}")

        domain_name = None
        try:
            domain = self.connection.identity.get_domain(project.domain_id)
            domain_name = domain.name
        except Exception as e:
            logger.warning(f"Get domain failed for {project.domain_id} error: {e}")

        lb_count = project_lb_dict.get(project.id)

        project_ret = Project(
            project_id=project.id,
            project_name=project.name,
            domain_id=project.domain_id,
            domain_name=domain_name,
            enabled=project.is_enabled,
            parent_id=project.parent_id,
            usage_instance=usage_instance, 
            quota_instance=quota_instance,
            usage_ram=usage_ram,
            quota_ram=quota_ram, 
            usage_vcpu=usage_vcpu,
            quota_vcpu=quota_vcpu,
            usage_volume=usage_volume,
            quota_volume=quota_volume,
            usage_snapshot=usage_snapshot,
            quota_snapshot=quota_snapshot,
            usage_storage=usage_storage,
            quota_storage=quota_storage,
            load_balancer_count=lb_count
        )

        return project_ret



import logging


import pandas as pd


from osi_dump.exporter.network.network_exporter import NetworkExporter 

from osi_dump.model.network import Network 

from osi_dump import util

logger = logging.getLogger(__name__)


class ExcelNetworkExporter(NetworkExporter):
    def __init__(self, sheet_name: str, output_file: str):
        self.sheet_name = sheet_name
        self.output_file = output_file

    def export_networks(self, networks: list[Network]):
        df = pd.json_normalize([network.model_dump() for network in networks])
        
        df = util.expand_list_column(df, "subnets")

        logger.info(f"Exporting networks for {self.sheet_name}")
        try:
            util.export_data_excel(self.output_file, sheet_name=self.sheet_name, df=df)

            logger.info(f"Exported networks for {self.sheet_name}")
        except Exception as e:
            logger.warning(f"Exporting networks for {self.sheet_name} error: {e}")



from typing import Optional
from pydantic import BaseModel, ConfigDict



class Network(BaseModel):
    model_config = ConfigDict(strict=True)

    network_id: str
    project_id: str
    name: Optional[str]

    subnets: Optional[list[dict]] # {"id": str, "cidr": "str"}
    mtu: int  

    port_security_enabled: Optional[bool]

    network_type: str 
    segmentation_id: Optional[int]
    physical_network: Optional[str]

    status: Optional[str]
    shared: Optional[bool]

    created_at: Optional[str]
    updated_at: Optional[str] 


import logging

import concurrent

from openstack.connection import Connection
from openstack.identity.v3.role_assignment import RoleAssignment as OSRoleAssignment
from openstack.identity.v3.user import User as OSUser


from osi_dump.importer.role_assignment.role_assignment_importer import (
    RoleAssignmentImporter,
)
from osi_dump.model.role_assignment import RoleAssignment

from osi_dump.api.keystone import get_role_assignments, get_users

logger = logging.getLogger(__name__)


class OpenStackRoleAssignmentImporter(RoleAssignmentImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

        self.users: dict[str, dict] = {}
        self.roles = {}

    def _get_users(self):
        os_users = get_users(self.connection)

        for os_user in os_users:
            self.users[os_user["id"]] = os_user

    def _get_roles(self):
        os_roles = self.connection.identity.roles()

        for os_role in os_roles:
            self.roles[os_role.id] = os_role.name

    def import_role_assignments(self) -> list[RoleAssignment]:
        """Import role_assignments information from Openstack

        Raises:
            Exception: Raises exception if fetching role_assignment failed

        Returns:
            list[RoleAssignment]: _description_
        """

        logger.info(
            f"Importing role_assignments for {self.connection.auth['auth_url']}"
        )

        try:
            self._get_users()
        except Exception as e:
            logger.info(f"Getting user list failed {e}")

        try:
            self._get_roles()
        except Exception as e:
            logger.info(f"Getting role list failed {e}")

        try:
            osrole_assignments: list[OSRoleAssignment] = list(
                self.connection.identity.role_assignments()
            )

        except Exception as e:
            raise Exception(
                f"Can not fetch role_assignments for {self.connection.auth['auth_url']} {e}"
            ) from e

        role_assignments: list[RoleAssignment] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_role_assignment_info, role_assignment)
                for role_assignment in osrole_assignments
            ]
            for future in concurrent.futures.as_completed(futures):
                role_assignments.append(future.result())

        logger.info(f"Imported role_assignments for {self.connection.auth['auth_url']}")

        return role_assignments

    def _get_role_assignment_info(
        self, role_assignment: OSRoleAssignment
    ) -> RoleAssignment:

        user_id = None
        role_id = None

        try:
            user_id: str = role_assignment["user"]["id"]
        except Exception as e:
            logger.warning(f"Can not get user id: {e}")

        try:
            role_id = role_assignment["role"]["id"]
        except Exception as e:
            logger.warning(f"Can not get role id: {e}")

        user_name = None
        role_name = None

        password_expires_at = None
        options = None

        try:
            user_name = self.users[user_id]["name"]
        except Exception as e:
            logger.warning(f"Can not get user name: {e}")

        try:
            role_name = self.roles[role_id]
        except Exception as e:
            logger.warning(f"Can not get role name: {e}")

        try:
            password_expires_at = self.users[user_id]["password_expires_at"]
        except Exception as e: 
            logger.warning(f"Can not get password expires at: {e}")


        try:
            options = self.users[user_id]["options"]
        except Exception as e:
            logger.warning(f"Can not get option")

        role_assignment_ret = RoleAssignment(
            user_id=user_id,
            user_name=user_name,
            role_id=role_id,
            role_name=role_name,
            scope=role_assignment["scope"],
            password_expires_at=password_expires_at,
            options=options
        )

        return role_assignment_ret



import logging

import concurrent

from openstack.network.v2.network import Network as OSNetwork
from openstack.network.v2.subnet import Subnet as OSSubnet

from openstack.connection import Connection

from osi_dump.importer.network.network_importer import NetworkImporter 
from osi_dump.model.network import Network 


logger = logging.getLogger(__name__)


class OpenStackNetworkImporter(NetworkImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_networks(self) -> list[Network]:
        """Import networks information from Openstack

        Raises:
            Exception: Raises exception if fetching networks failed

        Returns:
            list[Network]: _description_
        """

        try:
            os_networks: list[OSNetwork] = list(self.connection.list_networks())
        except Exception as e:
            raise Exception(
                f"Can not fetch hypervisor for {self.connection.auth['auth_url']}"
            ) from e

        networks: list[Network] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._get_network_info, network)
                for network in os_networks
            ]
            for future in concurrent.futures.as_completed(futures):
                networks.append(future.result())

        logger.info(f"Imported networks for {self.connection.auth['auth_url']}")

        return networks


    def _get_network_info(
        self, network: OSNetwork, 
    ) -> Network:

        subnets = self._get_subnets_info(subnet_ids=network.subnet_ids)

        return Network(
            network_id=network.id,
            project_id=network.project_id, 
            name=network.name, 
            mtu=network.mtu, 
            port_security_enabled=network.is_port_security_enabled,
            network_type=network.provider_network_type, 
            physical_network=network.provider_physical_network,
            segmentation_id=network.provider_segmentation_id,
            status=network.status, 
            shared=network.is_shared,
            created_at=network.created_at,
            updated_at=network.updated_at,
            subnets=subnets
        )
    

    def _get_subnets_info(self, subnet_ids: list[str]) -> list[dict]: 
        subnets = []

        for subnet_id in subnet_ids: 
            os_subnet: OSSubnet = self.connection.get_subnet(name_or_id=subnet_id) 

            if not os_subnet: 
                continue 
                
            subnets.append({
                "id": os_subnet.id, 
                "cidr": os_subnet.cidr
            })

        return subnets



    


import logging

import concurrent

from openstack.connection import Connection

from openstack.load_balancer.v2.load_balancer import LoadBalancer as OSLoadBalancer

from osi_dump.importer.load_balancer.load_balancer_importer import (
    LoadBalancerImporter,
)
from osi_dump.model.load_balancer import LoadBalancer

import osi_dump.api.octavia as octavia_api

logger = logging.getLogger(__name__)


class OpenStackLoadBalancerImporter(LoadBalancerImporter):
    def __init__(self, connection: Connection):
        self.connection = connection

    def import_load_balancers(self) -> list[LoadBalancer]:
        """Import load_balancers information from Openstack

        Raises:
            Exception: Raises exception if fetching load_balancer failed

        Returns:
            list[LoadBalancer]: _description_
        """

        logger.info(f"Importing load_balancers for {self.connection.auth['auth_url']}")

        try:
            osload_balancers: list[OSLoadBalancer] = octavia_api.get_load_balancers(
                connection=self.connection
            )
        except Exception as e:
            raise Exception(
                f"Can not fetch load_balancers for {self.connection.auth['auth_url']} {e}"
            ) from e

        load_balancers: list[LoadBalancer] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [] 
            
            for load_balancer in osload_balancers:
                logger.info("Importing load_balancer: %s", load_balancer["id"])
                if load_balancer["id"] != None:
                    futures.append(executor.submit(self._get_load_balancer_info, load_balancer))
            
            for future in concurrent.futures.as_completed(futures):
                result = future.result() 
                
                if result != None: 
                    load_balancers.append(result)

        logger.info(f"Imported load_balancers for {self.connection.auth['auth_url']}")

        return load_balancers

    def _get_load_balancer_info(self, load_balancer: OSLoadBalancer) -> LoadBalancer:

        lb_flavor_name = None 
        lb_flavor_description = None 

        try: 
            lb_flavor = octavia_api.get_load_balancer_flavor(
                connection=self.connection, 
                flavor_id=load_balancer["flavor_id"]
            )

            if lb_flavor:
                lb_flavor_name = lb_flavor["name"]
                lb_flavor_description = lb_flavor["description"]
            else: 
                raise Exception(f'No flavor id found for {load_balancer["id"]}')
        except Exception as e: 
            logger.warning(f"Get load balancer flavor failed {e}")

        try: 
            amphoraes = octavia_api.get_amphoraes(
                connection=self.connection, load_balancer_id=load_balancer["id"]
            )

            for amphorae in amphoraes:
                flavor = self.connection.get_flavor_by_id(amphorae["compute_flavor"])
                amphorae["ram"] = flavor.ram
                amphorae["vcpus"] = flavor.vcpus
                amphorae["flavor_name"] = flavor.name 
                amphorae["flavor_description"] = flavor.description

            load_balancer_ret = LoadBalancer(
                id=load_balancer["id"],
                load_balancer_name=load_balancer["name"],
                operating_status=load_balancer["operating_status"],
                project_id=load_balancer["project_id"],
                provisioning_status=load_balancer["provisioning_status"],
                created_at=load_balancer["created_at"],
                updated_at=load_balancer["updated_at"],
                amphoraes=amphoraes,
                vip=load_balancer["vip_address"],
                flavor_name=lb_flavor_name, 
                flavor_description=lb_flavor_description
            )

            return load_balancer_ret
        except Exception as e: 
            logger.warning(f"Getting lb failed {e}")
            
            return None



from openstack.connection import Connection
from openstack.identity.v3.service import Service
from openstack.load_balancer.v2.load_balancer import LoadBalancer


import osi_dump.util.openstack_util as os_util


def get_load_balancer_flavor(connection: Connection, flavor_id: str) -> dict: 

    if not flavor_id: 
        return None

    octavia_endpoints = os_util.get_endpoints(
        connection=connection, service_type="load-balancer", interface="public"
    )

    response = None

    for endpoint in octavia_endpoints:
        try:
            url = f"{endpoint}/v2.0/lbaas/flavors/{flavor_id}"
            response = connection.session.get(url)
            if response.status_code == 200:
                break
        except Exception as e:
            print(e)

    if response is None:
        return None
    
    data = response.json()

    return data["flavor"]

def get_load_balancers(connection: Connection) -> list[LoadBalancer]:
    # octavia_endpoint = os_util.get_endpoint(
    #     connection=connection, service_type="load-balancer", interface="public"
    # )

    octavia_endpoints = os_util.get_endpoints(
        connection=connection, service_type="load-balancer", interface="public"
    )

    response = None

    for endpoint in octavia_endpoints:
        try:
            url = f"{endpoint}/v2.0/lbaas/loadbalancers"
            response = connection.session.get(url)
            if response.status_code == 200:
                break
        except Exception as e:
            print(e)

    if response is None:
        return None

    data = response.json()

    return data["loadbalancers"]


def get_amphoraes(connection: Connection, load_balancer_id: str) -> list[dict]:

    # octavia_endpoint = os_util.get_endpoint(
    #     connection=connection, service_type="load-balancer", interface="public"
    # )

    octavia_endpoints = os_util.get_endpoints(
        connection=connection, service_type="load-balancer", interface="public"
    )

    response = None

    for endpoint in octavia_endpoints:
        try:
            url = f"{endpoint}/v2/octavia/amphorae?load_balancer_id={load_balancer_id}&fields=status&fields=compute_id&fields=compute_flavor&fields=role"
            response = connection.session.get(url)
            if response.status_code == 200:
                break
        except Exception as e:
            print(e)

    if response is None:
        return None

    data = response.json()

    amphoraes = data["amphorae"]

    amphoraes = [dict(sorted(amphorae.items())) for amphorae in amphoraes]

    return amphoraes
